/*
 * Georgiana Diana Ciocirdel
 * Proiect PM 2015 - Guitar Tuner
 *
 */

#include "lcd.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define FALSE 0
#define TRUE 1

/******************************************************************************/
/* LED specific stuff. */
#define PORT_LED PORTD
#define BIT_LED PD7
#define REG_LED DDRD
/******************************************************************************/

/******************************************************************************/
/* Button specific stuff. */
#define PORT_BTN PORTD
#define PIN_BTN PIND
#define BIT_BTN PD6
#define REG_BTN DDRD
/******************************************************************************/

/******************************************************************************/
/* Bit specific stuff. */
#define toggle_bit(reg, bit)    reg ^=  (1 << bit)
#define set_bit(reg, bit)       reg |=  (1 << bit)
#define reset_bit(reg, bit)     reg &= ~(1 << bit)
/******************************************************************************/

/******************************************************************************/
/* LCD specific stuff. */
#define FST_LINE_FST_CHAR 0x00
#define SND_LINE_FST_CHAR 0x40
#define EMPTY_LINE "                "
/******************************************************************************/

/******************************************************************************/
/* Tuner specific stuff. */
#define PRESCALER 256

#define E  164.81
#define A  220.00
#define D  293.66
#define G  391.99
#define B  493.88
#define EH 659.26

/* Two times the standard guitar guitar tunings as per
 * http://en.wikipedia.org/wiki/Guitar_tunings#Standard */
float std_fr[] = {164.81, 220.00, 293.66, 391.99, 493.88, 659.26};
char *str_std_fr[] = {"E  ", "A  ", "D  ", "G  ", "B ", "EH"};
char chosen_string = 0;

/******************************************************************************/
void timer1_init (void) {
    TCCR1B |= (1 << CS12) | /* Prescaler = 256 */
              (1 << WGM12); /* Clear on compare, TOP = OCR1A */
    TIMSK1 |= ( 1<< OCIE1A);
    OCR1A = 31250; /* TOP */
}

void LCD_clear (void) {
    LCD_printAt (FST_LINE_FST_CHAR, EMPTY_LINE);
    LCD_printAt (SND_LINE_FST_CHAR, EMPTY_LINE);
}

void LED_init (void) {
    REG_LED |= (1 << BIT_LED);
}

void button_init (void) {
    REG_BTN &= ~(1 << BIT_BTN);
    PORT_BTN |= (1 << BIT_BTN);
}

void input_init (void) {
    DDRB &= ~(1 << PB7);
    PCICR |= (1 << PCIE1);
    PCMSK1 |= (1 << PCINT15);
}

void wait_for (int millisecs) {
    int i;

    for (i = 0; i < millisecs; i++) {
        _delay_ms (1);
    }
}

void refresh_snd_line (void) {
    LCD_printAt (SND_LINE_FST_CHAR, EMPTY_LINE);
}

void startup_LCD (void) {
    int i;

    /* Display welcome message and toggle LED.*/
    LCD_printAt (FST_LINE_FST_CHAR, "Welcome to");
    LCD_printAt (SND_LINE_FST_CHAR, "Guitar Tuner");
    toggle_bit (PORT_LED, BIT_LED);
    wait_for (2000);
    toggle_bit (PORT_LED, BIT_LED);
    LCD_clear ();

    /* Display available strings. */
    char *strings = strdup ("");
    for (i = 0; i < 6; i++) {
        strcat (strings, str_std_fr[i]);
    }
    LCD_printAt (FST_LINE_FST_CHAR, strings);

    /* Display cursors below available strings. */
    refresh_snd_line ();
    LCD_printAt (FST_LINE_FST_CHAR, "");
}

int count_down_up;
int sum = 0;
int count = 0;

void move_cursor (void) {
    if ((PIN_BTN & (1 << BIT_BTN)) == 0) {
        chosen_string = (chosen_string + 1) % 6;
        refresh_snd_line ();
        LCD_printAt (FST_LINE_FST_CHAR + chosen_string * 3, "");
        toggle_bit (PORT_LED, BIT_LED);
        wait_for (300);
        toggle_bit (PORT_LED, BIT_LED);
        count_down_up = 0;
        sum = 0;
        count = 0;
    }
}

int main (void) {
    count_down_up = 0;

    /* Init LED. */
    LED_init ();

    /* Init button for input. */
    button_init ();

    /* Init LCD and show welcome message. */
    LCD_init ();
    startup_LCD ();

    /* Init input and timer. */
    input_init ();
    timer1_init ();
    sei();

    /* Infinite loop to run program. */
    for (;;) {
        move_cursor ();
    }

    return 0;
}

char buff[10];

ISR (PCINT1_vect) {
    if ((PINB & (1 << PB7)) == 0) {
        /* If there is a low-high transition */
        count_down_up++;
    }
}

/* Timer interrupt. */
ISR (TIMER1_COMPA_vect) {
    sum += count_down_up;
    count++;

    if (count == 3) {
        refresh_snd_line ();
        toggle_bit (PORT_LED, BIT_LED);
        sum /= count;
        int diff = sum - std_fr[chosen_string];
        char c = '-';
        if (diff > 0) {
            c = '+';
        } else if (diff == 0) {
            c = '0';
        }
        int abs_diff = abs (diff);
        float percent = (float) abs_diff / (std_fr[chosen_string] * 2) * 100;
        int percent_i = (int) percent;
        itoa (percent_i, buff, 10);
        int n = strlen (buff);
        buff[n] = '%';;
        buff[n + 1] = c;
        buff[n + 2] = '\0';
        LCD_printAt (SND_LINE_FST_CHAR, buff);
        LCD_printAt (FST_LINE_FST_CHAR + chosen_string * 3, "");

        count = 0;
        sum = 0;
    }

    count_down_up = 0;
}
